#include <iostream>
#include <map>
#include "headers/Yerbacon.hpp"
#include <future>
#include <filesystem>
using namespace std;

#include "headers/misc.hpp"
#include "headers/arguments.hpp"
#include "headers/transpiler/Target.hpp"

int main(int argc, char* argv[]) {
    setlocale(LC_ALL, "");
    string target = "lua";
    bool printResult = false,
         parallel = false,
         newLines = true,
         text_provided = false;
    using unit_result = pair<string, optional<Yerbacon::Exception>>;
    using unit = future<unit_result>;
    map<string_view, unit> Units;
    #ifdef EMSCRIPTEN
    is_node = MAIN_THREAD_EM_ASM_INT({ return ENVIRONMENT_IS_NODE; });
    #endif
    for (signed int i = 1; i < argc; ++i)
    {
        const string_view currentArgument (argv[i]);
        if (currentArgument == ArgumentShort("printresult")) printResult = true;
        else if (currentArgument == ArgumentAssignable("target")) {
            const string_view value = ArgumentAssignable::getValueFor(currentArgument);
            if (not value.empty()) target = value;
            else Yerbacon::fail("No target was provided.");
        }
        else if (currentArgument == Argument("parallel")) parallel = true;
        else if (currentArgument == ArgumentAssignable("newlines")) {
            const string_view enabled = ArgumentAssignable::getValueFor(currentArgument);
            if (enabled == "off") {
                newLines = false;
            } else if (enabled == "on") {
                newLines = true;
            } else goto invalid_argument;
        }
        else if (currentArgument == ArgumentShort("text")) printResult = text_provided = true;
        else if (text_provided || currentArgument.ends_with(".ybcon")) {
            #ifdef _OPENMP
            if (not parallel) omp_set_num_threads(1);
            #endif
            optional file_path = make_optional<filesystem::path>();
            if (not text_provided) {
                file_path = filesystem::path(currentArgument);
            #ifdef EMSCRIPTEN
                if (currentArgument.size() >= 2 and currentArgument[1] == ':') {
                    auto path = string(currentArgument).erase(0, 2);
                    replace(path.begin(), path.end(), '\\', '/');
                    file_path = path;
                }
                const string directory = filesystem::path(file_path.value()).remove_filename();
                for (unsigned int character = 1; character < directory.size(); ++character) {
                    const string sub_directory = directory.substr(0, character);
                    if (is_node and (directory[character] == '/') and not filesystem::is_directory(sub_directory)) {
                        MAIN_THREAD_EM_ASM({
                            const sub_directory = UTF8ToString($0);
                            if (!FS.isMountpoint(sub_directory)) {
                                try {
                                    FS.mkdir(sub_directory);
                                    FS.mount(NODEFS, {root: sub_directory}, sub_directory);
                                } catch (exception) {}
                            };
                         }, sub_directory.c_str());
                    }
                }
            #endif
            }
            Units.insert_or_assign(currentArgument, async(not parallel ? launch::deferred : launch::async, [&, currentArgument, file_path]() {
                unit_result resultingPair;
                try {
                    resultingPair.first = Target::forName(target, newLines)->transpileWithTree(
                        parseString(text_provided ? currentArgument : getFileContent(file_path->string()))
                    );
                    if (not text_provided) outputFileContent(filesystem::path(file_path.value()).replace_extension(target).string(), resultingPair.first);
                } catch (const Yerbacon::Exception& error) {
                    resultingPair.first = file_path.has_value() ? file_path->filename().string() : string();
                    resultingPair.second.emplace(error);
                }
                return resultingPair;
            }));
        } else {
            if (argc == 2) {
                if (currentArgument == Argument("version")) {
                    cout << Yerbacon::getVersion();
                } else if (currentArgument == Argument("buildInfo")) {
                    cout << Yerbacon::getBuildInfo();
                } else goto invalid_argument;
                cout << '\n'; exit(EXIT_SUCCESS);
            } else invalid_argument: Yerbacon::fail({"\"", currentArgument.data(), "\" is not a valid argument."});
        }
    }
    if (!Units.empty()) {
        if (printResult) cout << "~~~~[Yerbacon compilation result]~~~~\n\n";
        for (auto entry_iterator = Units.rbegin(); entry_iterator != Units.rend(); ++entry_iterator) {
            const pair result = entry_iterator->second.get();
            const bool is_exception = result.second.has_value();
            if (not is_exception) {
                if (printResult && !(result.first.empty() || all_of(result.first.begin(), result.first.end(), [](const char& character){
                    return isspace(character);
                }))) {
                    cout << result.first << '\n';
                }
            } else {
                cout << "Compilation"; if (not text_provided) cout << " of " << result.first; cout << " has failed with the following error:" << endl;
                cerr << result.second.value().what() << '\n';
                exit(EXIT_FAILURE);
            }
        }
    } else cout << (!text_provided ? "No valid file provided" : "No text was provided") << ".\n";
    exit(EXIT_SUCCESS);
}